"""Module grouping tests for the pydov.util.query module."""
from itertools import permutations

import numpy as np
import pandas as pd
import pytest
from owslib.etree import etree

from pydov.util.dovutil import build_dov_url
from pydov.util.query import Join, PropertyInList
from tests.abstract import clean_xml


class TestPropertyInList(object):
    """Test the PropertyInList query expression."""
    def test(self):
        """Test the PropertyInList expression with a standard list.

        Test whether the generated query is correct.

        """
        l = ['a', 'b', 'c']

        query = PropertyInList('methode', l)
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}Or'
        assert len(list(xml)) == 3

        for f in xml:
            assert f.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

            propertyname = f.find('./{http://www.opengis.net/ogc}PropertyName')
            assert propertyname.text == 'methode'

            literal = f.find('./{http://www.opengis.net/ogc}Literal')
            assert literal.text in l

            l.remove(literal.text)

        assert len(l) == 0

    def test_stable(self):
        """Test the PropertyInList expression with a standard list.

        Test whether the generated query is correct and stable.

        """
        l = ['a', 'b', 'c']

        for p in permutations(l):
            query = PropertyInList('methode', list(p))
            xml = query.toXML()

            assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
                '<ogc:Or><ogc:PropertyIsEqualTo><ogc:PropertyName>methode</ogc'
                ':PropertyName><ogc:Literal>a</ogc:Literal></ogc'
                ':PropertyIsEqualTo><ogc:PropertyIsEqualTo><ogc:PropertyName'
                '>methode</ogc:PropertyName><ogc:Literal>b</ogc:Literal></ogc'
                ':PropertyIsEqualTo><ogc:PropertyIsEqualTo><ogc:PropertyName'
                '>methode</ogc:PropertyName><ogc:Literal>c</ogc:Literal></ogc'
                ':PropertyIsEqualTo></ogc:Or>')

    def test_duplicate(self):
        """Test the PropertyInList expression with a list containing
        duplicates.

        Test whether the generated query is correct and does not contain the
        duplicate entry twice.

        """
        l = ['a', 'a', 'b', 'c']
        l_output = ['a', 'b', 'c']

        query = PropertyInList('methode', l)
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}Or'
        assert len(list(xml)) == 3

        for f in xml:
            assert f.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

            propertyname = f.find('./{http://www.opengis.net/ogc}PropertyName')
            assert propertyname.text == 'methode'

            literal = f.find('./{http://www.opengis.net/ogc}Literal')
            assert literal.text in l

            l_output.remove(literal.text)

        assert len(l_output) == 0

    def test_list_single(self):
        """Test the PropertyInList expression with a list containing
        a single item.

        Test whether the generated query is correct and does contain only a
        single PropertyIsEqualTo.

        """
        l = ['a']

        query = PropertyInList('methode', l)
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

        propertyname = xml.find('./{http://www.opengis.net/ogc}PropertyName')
        assert propertyname.text == 'methode'

        literal = xml.find('./{http://www.opengis.net/ogc}Literal')
        assert literal.text in l

        l.remove(literal.text)
        assert len(l) == 0

    def test_list_single_duplicate(self):
        """Test the PropertyInList expression with a list containing
        a single duplicated item.

        Test whether the generated query is correct and does contain only a
        single PropertyIsEqualTo.

        """
        l = ['a', 'a']
        l_output = ['a']

        query = PropertyInList('methode', l)
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

        propertyname = xml.find('./{http://www.opengis.net/ogc}PropertyName')
        assert propertyname.text == 'methode'

        literal = xml.find('./{http://www.opengis.net/ogc}Literal')
        assert literal.text in l_output

        l_output.remove(literal.text)
        assert len(l_output) == 0

    def test_emptylist(self):
        """Test the PropertyInList expression with an empty list.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            l = []
            PropertyInList('methode', l)

    def test_nolist(self):
        """Test the PropertyInList expression with a string instead of a list.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            l = 'goed'
            PropertyInList('betrouwbaarheid', l)


class TestJoin(object):
    """Test the Join query expression."""
    def test(self):
        """Test the Join expression with a standard dataframe.

        Test whether the generated query is correct.

        """
        l = [build_dov_url('data/boring/1986-068853'),
             build_dov_url('data/boring/1986-068843'),
             build_dov_url('data/boring/1980-068861')]

        df = pd.DataFrame({
            'pkey_boring': pd.Series(l),
            'diepte_tot_m': pd.Series([10, 20, 30])
        })

        query = Join(df, 'pkey_boring')
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}Or'
        assert len(list(xml)) == 3

        for f in xml:
            assert f.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

            propertyname = f.find('./{http://www.opengis.net/ogc}PropertyName')
            assert propertyname.text == 'pkey_boring'

            literal = f.find('./{http://www.opengis.net/ogc}Literal')
            assert literal.text in l

            l.remove(literal.text)

        assert len(l) == 0

    def test_duplicate(self):
        """Test the Join expression with a column containing
        duplicates.

        Test whether the generated query is correct and does not contain the
        duplicate entry twice.

        """
        l = [build_dov_url('data/boring/1986-068853'),
             build_dov_url('data/boring/1986-068853'),
             build_dov_url('data/boring/1980-068861')]

        l_output = [build_dov_url('data/boring/1986-068853'),
                    build_dov_url('data/boring/1980-068861')]

        df = pd.DataFrame({
            'pkey_boring': pd.Series(l),
            'diepte_tot_m': pd.Series([10, 20, 30])
        })

        query = Join(df, 'pkey_boring')
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}Or'
        assert len(list(xml)) == 2

        for f in xml:
            assert f.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

            propertyname = f.find('./{http://www.opengis.net/ogc}PropertyName')
            assert propertyname.text == 'pkey_boring'

            literal = f.find('./{http://www.opengis.net/ogc}Literal')
            assert literal.text in l

            l_output.remove(literal.text)

        assert len(l_output) == 0

    def test_wrongcolumn(self):
        """Test the Join expression with a join_column not available in the
        dataframe.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            l = [build_dov_url('data/boring/1986-068853'),
                 build_dov_url('data/boring/1986-068843'),
                 build_dov_url('data/boring/1980-068861')]

            df = pd.DataFrame({
                'pkey_boring': pd.Series(l),
                'diepte_tot_m': pd.Series([10, 20, 30])
            })

            Join(df, 'pkey_sondering')

    def test_single(self):
        """Test the Join expression with a dataframe containing a single row.

        Test whether the generated query is correct and does contain only a
        single PropertyIsEqualTo.

        """
        l = [build_dov_url('data/boring/1986-068853')]

        df = pd.DataFrame({
            'pkey_boring': pd.Series(l),
            'diepte_tot_m': pd.Series([10])
        })

        query = Join(df, 'pkey_boring')
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

        propertyname = xml.find('./{http://www.opengis.net/ogc}PropertyName')
        assert propertyname.text == 'pkey_boring'

        literal = xml.find('./{http://www.opengis.net/ogc}Literal')
        assert literal.text in l

        l.remove(literal.text)
        assert len(l) == 0

    def test_single_duplicate(self):
        """Test the Join expression with a dataframe containing two
        identical keys.

        Test whether the generated query is correct and does contain only a
        single PropertyIsEqualTo.

        """
        l = [build_dov_url('data/boring/1986-068853'),
             build_dov_url('data/boring/1986-068853')]
        l_output = [build_dov_url('data/boring/1986-068853')]

        df = pd.DataFrame({
            'pkey_boring': pd.Series(l),
            'diepte_tot_m': pd.Series([10, 20])
        })

        query = Join(df, 'pkey_boring')
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

        propertyname = xml.find('./{http://www.opengis.net/ogc}PropertyName')
        assert propertyname.text == 'pkey_boring'

        literal = xml.find('./{http://www.opengis.net/ogc}Literal')
        assert literal.text in l_output

        l_output.remove(literal.text)
        assert len(l_output) == 0

    def test_empty(self):
        """Test the Join expression with an empty dataframe.

        Test whether a ValueError is raised

        """
        df = pd.DataFrame({
            'pkey_boring': [np.nan, np.nan],
            'diepte_tot_m': pd.Series([10, 20])
        })

        with pytest.raises(ValueError):
            Join(df, 'pkey_boring')

    def test_on(self):
        """Test the Join expression with a standard dataframe and 'on'.

        Test whether the generated query is correct.

        """
        l = [build_dov_url('data/boring/1986-068853'),
             build_dov_url('data/boring/1986-068843'),
             build_dov_url('data/boring/1980-068861')]

        df = pd.DataFrame({
            'pkey_boring': pd.Series(l),
            'diepte_tot_m': pd.Series([10, 20, 30])
        })

        query = Join(df, on='pkey_boring')
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}Or'
        assert len(list(xml)) == 3

        for f in xml:
            assert f.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

            propertyname = f.find('./{http://www.opengis.net/ogc}PropertyName')
            assert propertyname.text == 'pkey_boring'

            literal = f.find('./{http://www.opengis.net/ogc}Literal')
            assert literal.text in l

            l.remove(literal.text)

        assert len(l) == 0

    def test_using(self):
        """Test the Join expression with a standard dataframe and 'on' and
        'using'.

        Test whether the generated query is correct.

        """
        l = [build_dov_url('data/boring/1986-068853'),
             build_dov_url('data/boring/1986-068843'),
             build_dov_url('data/boring/1980-068861')]

        df = pd.DataFrame({
            'boringfiche': pd.Series(l),
            'diepte_tot_m': pd.Series([10, 20, 30])
        })

        query = Join(df, on='pkey_boring', using='boringfiche')
        xml = query.toXML()

        assert xml.tag == '{http://www.opengis.net/ogc}Or'
        assert len(list(xml)) == 3

        for f in xml:
            assert f.tag == '{http://www.opengis.net/ogc}PropertyIsEqualTo'

            propertyname = f.find('./{http://www.opengis.net/ogc}PropertyName')
            assert propertyname.text == 'pkey_boring'

            literal = f.find('./{http://www.opengis.net/ogc}Literal')
            assert literal.text in l

            l.remove(literal.text)

        assert len(l) == 0



# -*- coding: utf-8 -*-
"""Module grouping network-related utilities and functions."""

import os
from queue import Empty, Queue
from threading import Thread

import requests
import urllib3
from requests.adapters import HTTPAdapter

import pydov

request_timeout = 300


def proxy_autoconfiguration():
    """Try proxy autoconfiguration via PAC.

    This function tries to autodetect the required proxy server using PAC, and
    sets the HTTP_PROXY and HTTPS_PROXY environment variables accordingly.

    These variables should subsequently be picked up by the requests sessions
    used by pydov and owslib.
    """
    def get_orig_proxy():
        """Get the proxy from current environment, if available.

        Returns
        -------
        tuple(str, str)
            The HTTP and HTTPS proxy respectively.
        """
        return (os.environ.get('HTTP_PROXY', None),
                os.environ.get('HTTPS_PROXY', None))

    def set_proxy_for_url(url):
        """Use PAC to discover the required proxy for the given URL and
        set the environment accordingly.

        Parameters
        ----------
        url : str
            The URL to pass to the PAC to determine the required proxy.
        """
        with pypac.pac_context_for_url(url):
            http_proxy = os.environ.get("HTTP_PROXY", "")
            https_proxy = os.environ.get("HTTPS_PROXY", "")

        os.environ['HTTP_PROXY'] = http_proxy
        os.environ['HTTPS_PROXY'] = https_proxy

    def revert_to_orig_proxy(orig_http_proxy, orig_https_proxy):
        """Revert the proxy environment to the given values.

        Parameters
        ----------
        orig_http_proxy : str
            Proxy server to use for HTTP, or None to disable.
        orig_https_proxy : str
            Proxy server to use for HTTPS, ot None to disable.
        """
        if orig_http_proxy is None:
            del os.environ['HTTP_PROXY']
        else:
            os.environ['HTTP_PROXY'] = orig_http_proxy

        if orig_https_proxy is None:
            del os.environ['HTTPS_PROXY']
        else:
            os.environ['HTTPS_PROXY'] = orig_https_proxy

    try:
        import pypac
    except ImportError:
        # do nothing if PAC not available
        pass
    else:
        # save original proxy from environment
        orig_http_proxy, orig_https_proxy = get_orig_proxy()

        from pydov.util.dovutil import build_dov_url
        dov_url = build_dov_url('/')
        public_url = 'https://pydov.readthedocs.io'

        # set proxy using PAC for DOV URL
        set_proxy_for_url(dov_url)
        try:
            # try if it works
            r = requests.get(dov_url)
            if not r.ok:
                raise RuntimeError
        except Exception:
            # fallback

            # set proxy using PAC for a public URL
            set_proxy_for_url(public_url)
            try:
                # try if it works
                r = requests.get(dov_url)
                if not r.ok:
                    raise RuntimeError
            except Exception:
                # if it does not, revert to original environment
                revert_to_orig_proxy(orig_http_proxy, orig_https_proxy)


class TimeoutHTTPAdapter(HTTPAdapter):
    """HTTPAdapter which adds a default timeout to requests. Allows timeout
    to be overridden on a per-request basis.
    """

    def __init__(self, *args, **kwargs):
        """Initialisation."""
        self.timeout = request_timeout
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        """Sends PreparedRequest object. Returns Response object.

        Parameters
        ----------
        request : requests.PreparedRequest
            The PreparedRequest being sent.

        Returns
        -------
        requests.Response
            The Response of the request.
        """
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


class SessionFactory:
    """Class for generating pydov configured requests Sessions. They are used
    to send HTTP requests using our user-agent and with added retry-logic.

    One global session is used for all requests, and additionally one
    session is used per thread executing XML requests in parallel.
    """

    @staticmethod
    def get_session():
        """Request a new session.

        Returns
        -------
        requests.Session
            pydov configured requests Session.
        """
        session = requests.Session()

        session.headers.update(
            {'user-agent': 'pydov/{}'.format(pydov.__version__)})

        try:
            retry = urllib3.util.Retry(
                total=10, connect=10, read=10, redirect=5, backoff_factor=1,
                allowed_methods=set(
                    ['HEAD', 'GET', 'POST', 'PUT', 'OPTIONS']))
        except TypeError:
            # urllib3 < 1.26.0 used method_whitelist instead
            retry = urllib3.util.Retry(
                total=10, connect=10, read=10, redirect=5, backoff_factor=1,
                method_whitelist=set(
                    ['HEAD', 'GET', 'POST', 'PUT', 'OPTIONS']))

        adapter = TimeoutHTTPAdapter(timeout=request_timeout,
                                     max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session


class LocalSessionThreadPool:
    """Thread pool of LocalSessionThreads used to perform XML I/O operations
    in parallel.
    """

    def __init__(self, workers=4):
        """Initialisation.

        Set up the pool and start all workers.

        Parameters
        ----------
        workers : int, optional
            Number of worker threads to use, defaults to 4.
        """
        self.workers = []
        self.input_queue = Queue(maxsize=100)
        self.result_queue = Queue()

        for i in range(workers):
            self.workers.append(LocalSessionThread(self.input_queue))

        self._start()

    def _start(self):
        """Start all worker threads. """
        for w in self.workers:
            w.start()

    def stop(self):
        """Stop all worker threads. """
        for w in self.workers:
            w.stop()

    def execute(self, fn, args):
        """Execute the given function with its arguments in a worker thread.

        This will add the job to the queue and will not wait for the result.
        Use join() to retrieve the result.

        Parameters
        ----------
        fn : function
            Function to execute.
        args : tuple
            Arguments that will be passed to the function.
        """
        r = WorkerResult()
        self.input_queue.put((fn, args, r))
        self.result_queue.put(r)

    def join(self):
        """Wait for all the jobs to be executed and return the results of all
        jobs in a list.

        Yields
        ------
        WorkerResult
            Results of the executed functions in the order they were
            submitted.
        """
        self.input_queue.join()
        self.stop()

        while not self.result_queue.empty():
            yield self.result_queue.get()


class WorkerResult:
    """Class for storing the result of a job execution in the result queue.

    This allows putting a result instance in the queue on job submission and
    fill in the result later when the job completes. This ensures the result
    output is in the same order as the jobs were submitted.
    """

    def __init__(self):
        """Initialisation. """
        self.result = None
        self.error = None

    def set_result(self, value):
        """Set the result of this job.

        Parameters
        ----------
        value : any
            The result of the execution of the job.
        """
        self.result = value

    def get_result(self):
        """Retrieve the result of this job.

        Returns
        -------
        any
            The result of the execution of the job.
        """
        return self.result

    def set_error(self, error):
        """Set the error, in case the jobs fails with an exception.

        Parameters
        ----------
        error : Exception
            The exception raised while executing this job.
        """
        self.error = error

    def get_error(self):
        """Retrieve the error, if any, of this job.

        Returns
        -------
        Exception
            The exception raised while executing this job.
        """
        return self.error


class LocalSessionThread(Thread):
    """Worker thread using a local Session to execute functions. """

    def __init__(self, input_queue):
        """Initialisation.

        Bind to the input queue and create a Session.

        Parameters
        ----------
        input_queue : queue.Queue
            Queue to poll for input, this should be in the form of a tuple with
            3 items: function to call, list with arguments and WorkerResult
            instance to store the output. The list with arguments will be
            automatically extended with the local Session instance.
        """
        super().__init__()
        self.input_queue = input_queue

        self.stopping = False
        self.session = SessionFactory.get_session()

    def stop(self):
        """Stop the worker thread at the next occasion. This can take up to
        500 ms. """
        self.stopping = True

    def run(self):
        """Executed while the thread is running. This is called implicitly
        when starting the thread. """
        while not self.stopping:
            try:
                fn, args, r = self.input_queue.get(timeout=0.5)
                args = list(args)
                args.append(self.session)

                try:
                    result = fn(*args)
                except BaseException as e:
                    r.set_error(e)
                else:
                    r.set_result(result)
                finally:
                    self.input_queue.task_done()
            except Empty:
                pass



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for bodemclassificatie."""
from pydov.types.fields import WfsField
from .abstract import AbstractDovType


class Bodemclassificatie(AbstractDovType):
    """Class representing the DOV data type for bodemclassificaties."""

    subtypes = []

    fields = [
        WfsField(name='pkey_bodemclassificatie',
                 source_field='Bodemclassificatiefiche',
                 datatype='string'),
        WfsField(name='pkey_bodemlocatie',
                 source_field='Bodemlocatiefiche',
                 datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='mv_mtaw', source_field='mv_mTAW', datatype='float'),
        WfsField(name='classificatietype',
                 source_field='Classificatietype', datatype='string'),
        WfsField(name='bodemtype', source_field='Bodemtype',
                 datatype='string'),
        WfsField(name='auteurs', source_field='Auteurs', datatype='string')
    ]

    pkey_fieldname = 'Bodemclassificatiefiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Bodemclassificatie, being a URI of the form
            `https://www.dov.vlaanderen.be/data/**classificatie/<id>`.

        """
        super().__init__('bodemclassificatie', pkey)



"""Module grouping tests for the sondering search module."""
import datetime

import pandas as pd
from owslib.fes import PropertyIsEqualTo

from pydov.search.sondering import SonderingSearch
from pydov.types.sondering import Sondering
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/sondering/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/sondering/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/sondering/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/sondering/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/sondering/feature.xml'
location_dov_xml = 'tests/data/types/sondering/sondering.xml'
location_xsd_base = 'tests/data/types/sondering/xsd_*.xml'


class TestSonderingSearch(AbstractTestSearch):

    search_instance = SonderingSearch()
    datatype_class = Sondering

    valid_query_single = PropertyIsEqualTo(propertyname='sondeernummer',
                                           literal='GEO-61/3075-S1')

    inexistent_field = 'onbestaand'
    wfs_field = 'sondeernummer'
    xml_field = 'gw_meting'

    valid_returnfields = (
        'pkey_sondering', 'sondeernummer', 'diepte_sondering_tot',
                          'datum_aanvang')
    valid_returnfields_subtype = (
        'pkey_sondering', 'sondeernummer', 'lengte', 'qc', 'Qt')
    valid_returnfields_extra = ('pkey_sondering', 'conus')

    df_default_columns = [
        'pkey_sondering', 'sondeernummer', 'x', 'y', 'mv_mtaw',
        'start_sondering_mtaw', 'diepte_sondering_van',
        'diepte_sondering_tot', 'datum_aanvang', 'uitvoerder',
        'sondeermethode', 'apparaat', 'datum_gw_meting',
        'diepte_gw_m', 'lengte', 'diepte', 'qc', 'Qt', 'fs', 'u', 'i']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.datum_aanvang.unique()[0] == datetime.date(2002, 12, 17)

        assert pd.Timestamp(
            df.datum_gw_meting.unique()[0]).to_pydatetime() == \
            datetime.datetime(2002, 12, 17, 14, 30, 0, 0)

    def test_search_nan(self, mp_wfs, mp_get_schema,
                        mp_remote_describefeaturetype, mp_remote_md,
                        mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        assert df.Qt.hasnans

    def test_search_xmlresolving(self, mp_get_schema,
                                 mp_remote_describefeaturetype,
                                 mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with return fields from XML but not from a
        subtype.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_sondering', 'sondeernummer', 'diepte_gw_m'))

        assert df.diepte_gw_m[0] == 3.60



"""Module grouping tests for the bodemlocatie search module."""
import datetime

from owslib.fes import PropertyIsEqualTo

from pydov.search.bodemlocatie import BodemlocatieSearch
from pydov.types.bodemlocatie import Bodemlocatie
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemlocatie/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemlocatie/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemlocatie/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemlocatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemlocatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemlocatie/bodemlocatie.xml'
location_xsd_base = 'tests/data/types/bodemlocatie/xsd_*.xml'


class TestBodemlocatieSearch(AbstractTestSearch):

    search_instance = BodemlocatieSearch()
    datatype_class = Bodemlocatie

    valid_query_single = PropertyIsEqualTo(propertyname='naam',
                                           literal='STARC_4')

    inexistent_field = 'onbestaand'
    wfs_field = 'naam'
    xml_field = 'invoerdatum'

    valid_returnfields = ('pkey_bodemlocatie', 'naam', 'waarnemingsdatum')
    valid_returnfields_subtype = (
        'pkey_bodemlocatie',
        'naam',
        'educatieve_waarde')
    valid_returnfields_extra = ('pkey_bodemlocatie', 'naam', 'erfgoed')

    df_default_columns = ['pkey_bodemlocatie', 'pkey_bodemsite',
                          'naam', 'type', 'waarnemingsdatum', 'doel', 'x', 'y',
                          'mv_mtaw', 'erfgoed', 'bodemstreek',
                          'invoerdatum',
                          'educatieve_waarde']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.waarnemingsdatum.unique()[0] == datetime.date(2011, 9, 16)



# -*- encoding: utf-8 -*-
import datetime
import gzip
import os
import time

import pandas as pd
import pytest
from owslib.fes import PropertyIsEqualTo

from pydov.search.boring import BoringSearch
from pydov.search.interpretaties import LithologischeBeschrijvingenSearch
from pydov.util.dovutil import build_dov_url
from pydov.util.errors import XmlParseWarning
from tests.abstract import ServiceCheck
from tests.test_search_itp_lithologischebeschrijvingen import (
    location_fc_featurecatalogue, location_md_metadata,
    location_wfs_describefeaturetype, location_wfs_getfeature)

location_dov_xml = 'tests/data/encoding/invalidcharacters.xml'


class TestEncoding(object):
    """Class grouping tests related to encoding issues."""

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_search(self, nocache):
        """Test the search method with strange character in the output.

        Test whether the output has the correct encoding.

        Parameters
        ----------
        nocache : pytest.fixture
            Fixture to disable caching.

        """
        boringsearch = BoringSearch()
        query = PropertyIsEqualTo(
            propertyname='pkey_boring',
            literal=build_dov_url('data/boring/1928-031159'))

        df = boringsearch.search(query=query,
                                 return_fields=('pkey_boring', 'uitvoerder'))

        assert df.uitvoerder[0] == u'Societé Belge des Bétons'

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('plaintext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['plaintext_cache'])
    def test_search_plaintext_cache(self, plaintext_cache):
        """Test the search method with strange character in the output.

        Test whether the output has the correct encoding, both with and
        without using the cache.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        boringsearch = BoringSearch()
        query = PropertyIsEqualTo(
            propertyname='pkey_boring',
            literal=build_dov_url('data/boring/1928-031159'))

        df = boringsearch.search(query=query,
                                 return_fields=('pkey_boring', 'uitvoerder',
                                                'mv_mtaw'))

        assert df.uitvoerder[0] == u'Societé Belge des Bétons'

        assert os.path.exists(os.path.join(
            plaintext_cache.cachedir, 'boring', '1928-031159.xml'))

        df = boringsearch.search(query=query,
                                 return_fields=('pkey_boring', 'uitvoerder',
                                                'mv_mtaw'))

        assert df.uitvoerder[0] == u'Societé Belge des Bétons'

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('gziptext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['gziptext_cache'])
    def test_search_gziptext_cache(self, gziptext_cache):
        """Test the search method with strange character in the output.

        Test whether the output has the correct encoding, both with and
        without using the cache.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        boringsearch = BoringSearch()
        query = PropertyIsEqualTo(
            propertyname='pkey_boring',
            literal=build_dov_url('data/boring/1928-031159'))

        df = boringsearch.search(query=query,
                                 return_fields=('pkey_boring', 'uitvoerder',
                                                'mv_mtaw'))

        assert df.uitvoerder[0] == u'Societé Belge des Bétons'

        assert os.path.exists(os.path.join(
            gziptext_cache.cachedir, 'boring', '1928-031159.xml.gz'))

        df = boringsearch.search(query=query,
                                 return_fields=('pkey_boring', 'uitvoerder',
                                                'mv_mtaw'))

        assert df.uitvoerder[0] == u'Societé Belge des Bétons'

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('plaintext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['plaintext_cache'])
    def test_caching_plaintext(self, plaintext_cache):
        """Test the caching of an XML containing strange characters.

        Test whether the data is saved in the cache.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '1995-056089.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        plaintext_cache.get(
            build_dov_url('data/boring/1995-056089.xml')),
        assert os.path.exists(cached_file)

        with open(cached_file, 'r', encoding='utf-8') as cf:
            cached_data = cf.read()
            assert cached_data != ""

        first_download_time = os.path.getmtime(cached_file)

        time.sleep(0.5)
        plaintext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))
        # assure we didn't redownload the file:
        assert os.path.getmtime(cached_file) == first_download_time

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('gziptext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['gziptext_cache'])
    def test_caching_gziptext(self, gziptext_cache):
        """Test the caching of an XML containing strange characters.

        Test whether the data is saved in the cache.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '1995-056089.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        gziptext_cache.get(
            build_dov_url('data/boring/1995-056089.xml')),
        assert os.path.exists(cached_file)

        with gzip.open(cached_file, 'rb') as cf:
            cached_data = cf.read().decode('utf-8')
            assert cached_data != ""

        first_download_time = os.path.getmtime(cached_file)

        time.sleep(0.5)
        gziptext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))
        # assure we didn't redownload the file:
        assert os.path.getmtime(cached_file) == first_download_time

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('plaintext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['plaintext_cache'])
    def test_save_content_plaintext(self, plaintext_cache):
        """Test the caching of an XML containing strange characters.

        Test if the contents of the saved document are the same as the
        original data.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '1995-056089.xml')

        plaintext_cache.remove()
        assert not os.path.exists(cached_file)

        ref_data = plaintext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))
        assert os.path.exists(cached_file)

        with open(cached_file, 'r', encoding='utf-8') as cached:
            cached_data = cached.read().encode('utf-8')

        assert cached_data == ref_data

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('gziptext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['gziptext_cache'])
    def test_save_content_gziptext(self, gziptext_cache):
        """Test the caching of an XML containing strange characters.

        Test if the contents of the saved document are the same as the
        original data.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '1995-056089.xml.gz')

        gziptext_cache.remove()
        assert not os.path.exists(cached_file)

        ref_data = gziptext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))
        assert os.path.exists(cached_file)

        with gzip.open(cached_file, 'rb') as cached:
            cached_data = cached.read()

        assert cached_data == ref_data

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('plaintext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['plaintext_cache'])
    def test_reuse_content_plaintext(self, plaintext_cache):
        """Test the caching of an XML containing strange characters.

        Test if the contents returned by the cache are the same as the
        original data.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '1995-056089.xml')

        plaintext_cache.remove()
        assert not os.path.exists(cached_file)

        ref_data = plaintext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))
        assert os.path.exists(cached_file)

        cached_data = plaintext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))

        assert cached_data == ref_data

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('gziptext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['gziptext_cache'])
    def test_reuse_content_gziptext(self, gziptext_cache):
        """Test the caching of an XML containing strange characters.

        Test if the contents returned by the cache are the same as the
        original data.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '1995-056089.xml.gz')

        gziptext_cache.remove()
        assert not os.path.exists(cached_file)

        ref_data = gziptext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))
        assert os.path.exists(cached_file)

        cached_data = gziptext_cache.get(
            build_dov_url('data/boring/1995-056089.xml'))

        assert cached_data == ref_data

    def test_search_invalidxml_single(
            self, mp_wfs, mp_remote_describefeaturetype, mp_get_schema,
            mp_remote_md, mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml,
            nocache):
        """Test the search method when the XML is invalid.

        If lxml is installed, the XML should parse regardless of invalid
        characters. If lxml is not installed, the dataframe should be
        returned with a warning that it will be incomplete.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.
        nocache : pytest.fixture
            Fixture to disable caching.

        """
        lithosearch = LithologischeBeschrijvingenSearch()
        query = PropertyIsEqualTo(
            propertyname='pkey_interpretatie',
            literal=build_dov_url('data/interpretatie/1987-070909'))

        try:
            import lxml
            have_lxml = True
        except ImportError:
            have_lxml = False

        if have_lxml:
            df = lithosearch.search(query=query)
            assert df.beschrijving[1].startswith(u'homogene groene zanden')
        else:
            with pytest.warns(XmlParseWarning):
                df = lithosearch.search(query=query)
                assert len(df) == 1
                assert pd.isnull(df.diepte_laag_van[0])
                assert pd.isnull(df.diepte_laag_tot[0])
                assert pd.isnull(df.beschrijving[0])



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV borehole data."""
from ..types.boring import Boring
from .abstract import AbstractSearch


class BoringSearch(AbstractSearch):
    """Search class to retrieve information about boreholes (Boring)."""

    def __init__(self, objecttype=Boring):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Boring type.
            Optional: defaults to the Boring type containing the fields
            described in the documentation.

        """
        super(BoringSearch, self).__init__('dov-pub:Boringen', objecttype)



"""Script to update the testdata based on DOV webservices."""
import os
import sys

from owslib.etree import etree

from pydov.types.bodemlocatie import Bodemlocatie
from pydov.types.bodemdiepteinterval import Bodemdiepteinterval
from pydov.types.bodemmonster import Bodemmonster
from pydov.types.bodemobservatie import Bodemobservatie
from pydov.types.bodemsite import Bodemsite
from pydov.types.bodemclassificatie import Bodemclassificatie
from pydov.types.boring import Boring
from pydov.types.grondmonster import Grondmonster
from pydov.types.grondwaterfilter import GrondwaterFilter
from pydov.types.grondwatermonster import GrondwaterMonster
from pydov.types.grondwatervergunning import GrondwaterVergunning
from pydov.types.interpretaties import (FormeleStratigrafie,
                                        GecodeerdeLithologie,
                                        GeotechnischeCodering,
                                        HydrogeologischeStratigrafie,
                                        InformeleHydrogeologischeStratigrafie,
                                        InformeleStratigrafie,
                                        LithologischeBeschrijvingen,
                                        QuartairStratigrafie)
from pydov.types.sondering import Sondering
from pydov.util.dovutil import build_dov_url, get_remote_url
from pydov.util.net import LocalSessionThreadPool
from tests.abstract import ServiceCheck


def get_first_featuremember(wfs_response):
    tree = etree.fromstring(wfs_response.encode('utf-8'))

    feature_members = tree.find('.//{http://www.opengis.net/gml}'
                                'featureMembers')

    if feature_members is not None:
        for ft in feature_members:
            return etree.tostring(ft).decode('utf-8')


def update_file_real(filepath, url, process_fn=None, session=None):
    output = 'Updated {} ... '.format(filepath)
    filepath = os.path.join(os.path.dirname(__file__), filepath)
    try:
        data = get_remote_url(url, session)
        if isinstance(data, bytes):
            data = data.decode('utf-8')
    except Exception as e:
        output += ' FAILED:\n   {}.\n'.format(e)
    else:
        if not os.path.isdir(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))

        with open(filepath, 'wb') as f:
            if process_fn:
                data = process_fn(data)
            f.write(data.encode('utf-8'))
            output += ' OK.\n'

    return output
    # with lock:
    #     sys.stdout.write(output)


if __name__ == '__main__':
    if not ServiceCheck.service_ok():
        print('Not updating test data as services are unavailable.')
        sys.exit(1)

    pool = LocalSessionThreadPool()

    def update_file(filepath, url, process_fn=None):
        pool.execute(update_file_real, (filepath, url, process_fn))

    # types/boring
    update_file('types/boring/boring.xml',
                build_dov_url('data/boring/2004-103984.xml'))

    update_file(
        'types/boring/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=dov-pub:Boringen'
            '&maxFeatures=1&CQL_Filter=fiche=%27' +
            build_dov_url('data/boring/2004-103984%27')))

    update_file(
        'types/boring/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=dov-pub:Boringen'
            '&maxFeatures=1&CQL_Filter=fiche=%27' +
            build_dov_url('data/boring/2004-103984%27')),
        get_first_featuremember)

    update_file(
        'types/boring/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=c0cbd397-520f-4ee1-aca7'
            '-d70e271eeed6'))

    update_file(
        'types/boring/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=4e20bf9c-3a5c-42be-b5b6'
            '-bef6214d1fa7'))

    update_file(
        'types/boring/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/dov-pub/Boringen'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Boring.get_xsd_schemas():
        update_file(
            'types/boring/xsd_{}.xml'.format(xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/sondering

    update_file('types/sondering/sondering.xml',
                build_dov_url('data/sondering/2002-018435.xml'))

    update_file(
        'types/sondering/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=dov-pub'
            ':Sonderingen&maxFeatures=1&CQL_Filter=fiche=%27' +
            build_dov_url('data/sondering/2002-018435%27')))

    update_file(
        'types/sondering/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=dov-pub'
            ':Sonderingen&maxFeatures=1&CQL_Filter=fiche=%27' +
            build_dov_url('data/sondering/2002-018435%27')),
        get_first_featuremember)

    update_file(
        'types/sondering/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=bd539ba5-5f4d-4c43-9662'
            '-51c16caea351'))

    update_file(
        'types/sondering/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=b397faec-1b64-4854-8000'
            '-2375edb3b1a8'))

    update_file(
        'types/sondering/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/dov-pub/Sonderingen'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Sondering.get_xsd_schemas():
        update_file(
            'types/sondering/xsd_{}.xml'.format(xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/interpretaties/informele_stratigrafie

    update_file('types/interpretaties/informele_stratigrafie'
                '/informele_stratigrafie.xml',
                build_dov_url('data/interpretatie/1962-101692.xml'))

    update_file(
        'types/interpretaties/informele_stratigrafie'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':informele_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/1962-101692%27'))

    update_file(
        'types/interpretaties/informele_stratigrafie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':informele_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/1962-101692%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/informele_stratigrafie/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=b6c651f9-5972-4252-ae10-ad69ad08e78d'))

    update_file(
        'types/interpretaties/informele_stratigrafie/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=bd171ea4-2509-478d-a21c'
            '-c2728d3a9051'))

    update_file(
        'types/interpretaties/informele_stratigrafie/wfsdescribefeaturetype'
        '.xml', build_dov_url(
            'geoserver/interpretaties'
            '/informele_stratigrafie/ows?service=wfs&version=1.1.0&request'
            '=DescribeFeatureType'))

    for xsd_schema in InformeleStratigrafie.get_xsd_schemas():
        update_file(
            'types/interpretaties/informele_stratigrafie/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/interpretaties/formele_stratigrafie

    update_file('types/interpretaties/formele_stratigrafie'
                '/formele_stratigrafie.xml',
                build_dov_url('data/interpretatie/2011-249333.xml'))

    update_file(
        'types/interpretaties/formele_stratigrafie'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':formele_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2011-249333%27'))

    update_file(
        'types/interpretaties/formele_stratigrafie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':formele_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2011-249333%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/formele_stratigrafie/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=68405b5d-51e6-44d0-b634-b580bc2f9eb6'))

    update_file(
        'types/interpretaties/formele_stratigrafie/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=212af8cd-bffd-423c-9d2b'
            '-69c544ab3b04'))

    update_file(
        'types/interpretaties/formele_stratigrafie/wfsdescribefeaturetype'
        '.xml', build_dov_url(
            'geoserver/interpretaties'
            '/formele_stratigrafie/ows?service=wfs&version=1.1.0&request'
            '=DescribeFeatureType'))

    for xsd_schema in FormeleStratigrafie.get_xsd_schemas():
        update_file(
            'types/interpretaties/formele_stratigrafie/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/interpretaties/hydrogeologische_stratigrafie

    update_file('types/interpretaties/hydrogeologische_stratigrafie'
                '/hydrogeologische_stratigrafie.xml',
                build_dov_url('data/interpretatie/2001-186543.xml'))

    update_file(
        'types/interpretaties/hydrogeologische_stratigrafie'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':hydrogeologische_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2001-186543%27'))

    update_file(
        'types/interpretaties/hydrogeologische_stratigrafie'
        '/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':hydrogeologische_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data/'
            'interpretatie/2001-186543%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/hydrogeologische_stratigrafie/'
        'fc_featurecatalogue.xml', build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=b89e72de-35a9-4bca-8d0b-712d1e881ea6'))

    update_file(
        'types/interpretaties/hydrogeologische_stratigrafie/'
        'md_metadata.xml', build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=25c5d9fa-c2ba-4184-b796'
            '-fde790e73d39'))

    update_file(
        'types/interpretaties/hydrogeologische_stratigrafie/'
        'wfsdescribefeaturetype.xml', build_dov_url(
            'geoserver/interpretaties'
            '/hydrogeologische_stratigrafie/ows?service=wfs&version=1.1.0'
            '&request=DescribeFeatureType'))

    for xsd_schema in HydrogeologischeStratigrafie.get_xsd_schemas():
        update_file(
            'types/interpretaties/hydrogeologische_stratigrafie/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/interpretaties/lithologische_beschrijvingen

    update_file('types/interpretaties/lithologische_beschrijvingen'
                '/lithologische_beschrijvingen.xml',
                build_dov_url('data/interpretatie/1958-003925.xml'))

    update_file(
        'types/interpretaties/lithologische_beschrijvingen'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':lithologische_beschrijvingen&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/1958-003925%27'))

    update_file(
        'types/interpretaties/lithologische_beschrijvingen/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':lithologische_beschrijvingen&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/1958-003925%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/lithologische_beschrijvingen/'
        'fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=2450d592-29bc-4970-a89f-a7b14bd38dc2'))

    update_file(
        'types/interpretaties/lithologische_beschrijvingen/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=45b5610e-9a66-42bd-b920'
            '-af099e399f3b'))

    update_file(
        'types/interpretaties/lithologische_beschrijvingen/'
        'wfsdescribefeaturetype.xml', build_dov_url(
            'geoserver/interpretaties'
            '/lithologische_beschrijvingen/ows?service=wfs&version=1.1.0'
            '&request=DescribeFeatureType'))

    for xsd_schema in LithologischeBeschrijvingen.get_xsd_schemas():
        update_file(
            'types/interpretaties/lithologische_beschrijvingen/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/interpretaties/gecodeerde_lithologie

    update_file('types/interpretaties/gecodeerde_lithologie'
                '/gecodeerde_lithologie.xml',
                build_dov_url('data/interpretatie/2001-046845.xml'))

    update_file(
        'types/interpretaties/gecodeerde_lithologie'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':gecodeerde_lithologie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2001-046845%27'))

    update_file(
        'types/interpretaties/gecodeerde_lithologie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':gecodeerde_lithologie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2001-046845%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/gecodeerde_lithologie/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=0032241d-8920-415e-b1d8-fa0a48154904'))

    update_file(
        'types/interpretaties/gecodeerde_lithologie/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=35d630e4-9145-46f9-b7dc'
            '-da290a0adc55'))

    update_file(
        'types/interpretaties/gecodeerde_lithologie/wfsdescribefeaturetype'
        '.xml', build_dov_url(
            'geoserver/interpretaties'
            '/gecodeerde_lithologie/ows?service=wfs&version=1.1.0&request'
            '=DescribeFeatureType'))

    for xsd_schema in GecodeerdeLithologie.get_xsd_schemas():
        update_file(
            'types/interpretaties/gecodeerde_lithologie/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/interpretaties/geotechnische_codering

    update_file('types/interpretaties/geotechnische_codering'
                '/geotechnische_codering.xml',
                build_dov_url('data/interpretatie/2016-298511.xml'))

    update_file(
        'types/interpretaties/geotechnische_codering'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':geotechnische_coderingen&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2016-298511%27'))

    update_file(
        'types/interpretaties/geotechnische_codering/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':geotechnische_coderingen&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2016-298511%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/geotechnische_codering/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=85404aa6-2d88-46f6-ae5a-575aece71efd'))

    update_file(
        'types/interpretaties/geotechnische_codering/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=6a3dc5d4-0744-4d9c-85ce'
            '-da50913851cc'))

    update_file(
        'types/interpretaties/geotechnische_codering/wfsdescribefeaturetype'
        '.xml', build_dov_url(
            'geoserver/interpretaties'
            '/geotechnische_coderingen/ows?service=wfs&version=1.1.0&request'
            '=DescribeFeatureType'))

    for xsd_schema in GeotechnischeCodering.get_xsd_schemas():
        update_file(
            'types/interpretaties/geotechnische_codering/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/interpretaties/informele_hydrogeologische_stratigrafie

    update_file('types/interpretaties/informele_hydrogeologische_stratigrafie'
                '/informele_hydrogeologische_stratigrafie.xml',
                build_dov_url('data/interpretatie/2003-297774.xml'))

    update_file(
        'types/interpretaties/informele_hydrogeologische_stratigrafie'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':informele_hydrogeologische_stratigrafie&maxFeatures=1'
            '&CQL_Filter=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2003-297774%27'))

    update_file(
        'types/interpretaties/informele_hydrogeologische_stratigrafie'
        '/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':informele_hydrogeologische_stratigrafie&maxFeatures=1'
            '&CQL_Filter=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/2003-297774%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/informele_hydrogeologische_stratigrafie'
        '/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=69f71840-bd29-4b59-9b02-4e36aafaa041'))

    update_file(
        'types/interpretaties/informele_hydrogeologische_stratigrafie'
        '/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full'
            '&id=ca1d704a-cdee-4968-aa65-9c353863e4b1'))

    update_file(
        'types/interpretaties/informele_hydrogeologische_stratigrafie/'
        'wfsdescribefeaturetype.xml', build_dov_url(
            'geoserver/interpretaties'
            '/informele_hydrogeologische_stratigrafie/'
            'ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in InformeleHydrogeologischeStratigrafie.get_xsd_schemas():
        update_file(
            'types/interpretaties/informele_hydrogeologische_stratigrafie/'
            'xsd_%s.xml' % xsd_schema.split('/')[-1], xsd_schema)

    # types/grondwaterfilter

    update_file('types/grondwaterfilter/grondwaterfilter.xml',
                build_dov_url('data/filter/2003-004471.xml'))

    update_file('types/grondwaterfilter/wfsgetfeature.xml',
                build_dov_url('geoserver/ows?service=WFS'
                              '&version=1.1.0&request=GetFeature&typeName='
                              'gw_meetnetten:meetnetten&maxFeatures=1&'
                              'CQL_Filter=filterfiche=%27' + build_dov_url(
                                  'data/filter/2003-004471%27')))

    update_file('types/grondwaterfilter/feature.xml',
                build_dov_url('geoserver/ows?service=WFS'
                              '&version=1.1.0&request=GetFeature&typeName='
                              'gw_meetnetten:meetnetten&maxFeatures=1&'
                              'CQL_Filter=filterfiche=%27' + build_dov_url(
                                  'data/filter/2003-004471%27')),
                get_first_featuremember)

    update_file(
        'types/grondwaterfilter/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=b142965f-b2aa-429e-86ff'
            '-a7cb0e065d48'))

    update_file(
        'types/grondwaterfilter/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=6c39d716-aecc-4fbc-bac8'
            '-4f05a49a78d5'))

    update_file('types/grondwaterfilter/wfsdescribefeaturetype.xml',
                build_dov_url('geoserver/gw_meetnetten/'
                              'meetnetten/ows?service=wfs&version=1.1.0&'
                              'request=DescribeFeatureType'))

    for xsd_schema in GrondwaterFilter.get_xsd_schemas():
        update_file(
            'types/grondwaterfilter/xsd_%s.xml' % xsd_schema.split('/')[-1],
            xsd_schema)

    # types/grondwatermonster

    update_file('types/grondwatermonster/grondwatermonster.xml',
                build_dov_url('data/watermonster/2006-115684.xml'))

    update_file(
        'types/grondwatermonster/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'gw_meetnetten:grondwatermonsters&maxFeatures=1&'
            'CQL_Filter=grondwatermonsterfiche=%27' +
            build_dov_url('data/watermonster/2006-115684') +
            '%27'))

    update_file(
        'types/grondwatermonster/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'gw_meetnetten:grondwatermonsters&maxFeatures=1&'
            'CQL_Filter=grondwatermonsterfiche=%27' +
            build_dov_url('data/watermonster/2006-115684') +
            '%27'),
        get_first_featuremember)

    update_file(
        'types/grondwatermonster/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&'
            'id=639c9612-4bbb-4826-86fd-fec9afd49bf7'))

    update_file(
        'types/grondwatermonster/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&'
            'id=0b378716-39fb-4151-96c5-2021672f4762'))

    update_file(
        'types/grondwatermonster/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/gw_meetnetten/'
            'grondwatermonsters/ows?service=wfs&version=1.1.0&'
            'request=DescribeFeatureType'))

    for xsd_schema in GrondwaterMonster.get_xsd_schemas():
        update_file(
            'types/grondwatermonster/xsd_%s.xml' % xsd_schema.split('/')[-1],
            xsd_schema)

    # util/owsutil

    update_file(
        'util/owsutil/fc_featurecatalogue_notfound.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=badfc000-0000-0000-0000'
            '-badfc00badfc'))

    update_file('util/owsutil/wfscapabilities.xml',
                build_dov_url('geoserver/wfs?request'
                              '=getcapabilities&service=wfs&version=1.1.0'))

    # types/interpretaties/quartaire_stratigrafie

    update_file('types/interpretaties/quartaire_stratigrafie'
                '/quartaire_stratigrafie.xml',
                build_dov_url('data/interpretatie/1999-057087.xml'))

    update_file(
        'types/interpretaties/quartaire_stratigrafie'
        '/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':quartaire_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/1999-057087%27'))

    update_file(
        'types/interpretaties/quartaire_stratigrafie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=interpretaties'
            ':quartaire_stratigrafie&maxFeatures=1&CQL_Filter'
            '=Interpretatiefiche=%27') +
        build_dov_url(
            'data'
            '/interpretatie/1999-057087%27'),
        get_first_featuremember)

    update_file(
        'types/interpretaties/quartaire_stratigrafie/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=d40ef884-3278-45db-ad69-2c2a8c3981c3'))

    update_file(
        'types/interpretaties/quartaire_stratigrafie/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=8b204ed6-e44c-4567-bbe8'
            '-bd427eba082c'))

    update_file(
        'types/interpretaties/quartaire_stratigrafie/wfsdescribefeaturetype'
        '.xml', build_dov_url(
            'geoserver/interpretaties'
            '/quartaire_stratigrafie/ows?service=wfs&version=1.1.0&request'
            '=DescribeFeatureType'))

    for xsd_schema in QuartairStratigrafie.get_xsd_schemas():
        update_file(
            'types/interpretaties/quartaire_stratigrafie/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/grondmonster

    update_file('types/grondmonster/grondmonster.xml',
                build_dov_url('data/grondmonster/2018-211728.xml'))

    update_file(
        'types/grondmonster/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'boringen:grondmonsters&maxFeatures=1&CQL_Filter'
            '=grondmonsterfiche=%27' +
            build_dov_url(
                'data'
                '/grondmonster/2018-211728') +
            '%27'))

    update_file(
        'types/grondmonster/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'boringen:grondmonsters&maxFeatures=1&CQL_Filter'
            '=grondmonsterfiche=%27' +
            build_dov_url(
                'data'
                '/grondmonster/2018-211728') +
            '%27'),
        get_first_featuremember)

    update_file(
        'types/grondmonster/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=b9338fb5-fc9c-4229-858b-06a5fa3ee49d'))

    update_file(
        'types/grondmonster/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&'
            'id=6edeab46-2cfc-4aa2-ae03-307d772f34ae'))

    update_file(
        'types/grondmonster/wfsdescribefeaturetype'
        '.xml',
        build_dov_url('geoserver/boringen'
                      '/grondmonsters/ows?service=wfs&version=1.1.0&request'
                      '=DescribeFeatureType'))

    for xsd_schema in Grondmonster.get_xsd_schemas():
        update_file(
            'types/grondmonster/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/bodemlocatie
    update_file('types/bodemlocatie/bodemlocatie.xml',
                build_dov_url('data/bodemlocatie/2011-000002.xml'))

    update_file(
        'types/bodemlocatie/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemlocaties'
            '&maxFeatures=1&CQL_Filter=Bodemlocatiefiche=%27' +
            build_dov_url('data/bodemlocatie/2011-000002%27')))

    update_file(
        'types/bodemlocatie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemlocaties'
            '&maxFeatures=1&CQL_Filter=Bodemlocatiefiche=%27' +
            build_dov_url('data/bodemlocatie/2011-000002%27')),
        get_first_featuremember)

    update_file(
        'types/bodemlocatie/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=89d4f9a1-0474-4ade-b30f-442c31d17dc6'))

    update_file(
        'types/bodemlocatie/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=3f507fd9-24c0-40ab-9328-29f0dff571fe'))

    update_file(
        'types/bodemlocatie/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/bodem/bodemlocaties'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Bodemlocatie.get_xsd_schemas():
        update_file(
            'types/bodemlocatie/xsd_{}.xml'.format(xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/bodemdiepteinterval
    update_file(
        'types/bodemdiepteinterval/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'bodem:bodemdiepteintervallen&maxFeatures=1&'
            'CQL_Filter=Diepteintervalfiche=%27' +
            build_dov_url('data/bodemdiepteinterval/2018-000004%27')))

    update_file(
        'types/bodemdiepteinterval/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'bodem:bodemdiepteintervallen&maxFeatures=1&'
            'CQL_Filter=Diepteintervalfiche=%27' +
            build_dov_url('data/bodemdiepteinterval/2018-000004%27')),
        get_first_featuremember)

    update_file(
        'types/bodemdiepteinterval/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=859ad05d-d6fc-4850-b040-1bdac3641ff5'))

    update_file(
        'types/bodemdiepteinterval/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=96a1127f-d7d0-4a33-b24c-04d982b63dce'))

    update_file(
        'types/bodemdiepteinterval/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/bodem/bodemdiepteintervallen'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Bodemdiepteinterval.get_xsd_schemas():
        update_file(
            'types/bodemdiepteinterval/xsd_{}.xml'.format(
                xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/bodemobservatie
    update_file('types/bodemobservatie/bodemobservatie.xml',
                build_dov_url('data/bodemobservatie/2019-001221.xml'))

    update_file(
        'types/bodemobservatie/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemobservaties'
            '&maxFeatures=1&CQL_Filter=Bodemobservatiefiche=%27' +
            build_dov_url('data/bodemobservatie/2019-001221%27')))

    update_file(
        'types/bodemobservatie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemobservaties'
            '&maxFeatures=1&CQL_Filter=Bodemobservatiefiche=%27' +
            build_dov_url('data/bodemobservatie/2019-001221%27')),
        get_first_featuremember)

    update_file(
        'types/bodemobservatie/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=44df1272-6b57-471b-9f7a-2dc82f760137'))

    update_file(
        'types/bodemobservatie/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=dd327fef-62c7-4980-9788-9fac047a1553'))

    update_file(
        'types/bodemobservatie/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/bodem/bodemobservaties'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Bodemobservatie.get_xsd_schemas():
        update_file(
            'types/bodemobservatie/xsd_{}.xml'.format(
                xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/bodemmonster
    update_file('types/bodemmonster/bodemmonster.xml',
                build_dov_url('data/bodemmonster/2015-211807.xml'))

    update_file(
        'types/bodemmonster/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemmonsters'
            '&maxFeatures=1&CQL_Filter=Bodemmonsterfiche=%27' +
            build_dov_url('data/bodemmonster/2015-211807%27')))

    update_file(
        'types/bodemmonster/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemmonsters'
            '&maxFeatures=1&CQL_Filter=Bodemmonsterfiche=%27' +
            build_dov_url('data/bodemmonster/2015-211807%27')),
        get_first_featuremember)

    update_file(
        'types/bodemmonster/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=7d69c092-fa5a-4399-86ed-003877f5899e'))

    update_file(
        'types/bodemmonster/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=ff1902b2-7ba2-46be-ba8c-bfcf893444c2'))

    update_file(
        'types/bodemmonster/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/bodem/bodemmonsters'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Bodemmonster.get_xsd_schemas():
        update_file(
            'types/bodemmonster/xsd_{}.xml'.format(xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/bodemsite
    update_file('types/bodemsite/bodemsite.xml',
                build_dov_url('data/bodemsite/2013-000180.xml'))

    update_file(
        'types/bodemsite/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemsites'
            '&maxFeatures=1&CQL_Filter=Bodemsitefiche=%27' +
            build_dov_url('data/bodemsite/2013-000180%27')))

    update_file(
        'types/bodemsite/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName=bodem:bodemsites'
            '&maxFeatures=1&CQL_Filter=Bodemsitefiche=%27' +
            build_dov_url('data/bodemsite/2013-000180%27')),
        get_first_featuremember)

    update_file(
        'types/bodemsite/fc_featurecatalogue.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gfc'
            '&elementSetName=full&id=955d4bc8-9d78-4af6-9782-85698caae0aa'))

    update_file(
        'types/bodemsite/md_metadata.xml',
        build_dov_url(
            'geonetwork/srv/dut/csw'
            '?Service=CSW&Request=GetRecordById&Version=2.0.2'
            '&outputSchema=http://www.isotc211.org/2005/gmd'
            '&elementSetName=full&id=27142078-f0d0-46e3-b97e-2ffc2c6bdd41'))

    update_file(
        'types/bodemsite/wfsdescribefeaturetype.xml',
        build_dov_url(
            'geoserver/bodem/bodemsites'
            '/ows?service=wfs&version=1.1.0&request=DescribeFeatureType'))

    for xsd_schema in Bodemsite.get_xsd_schemas():
        update_file(
            'types/bodemsite/xsd_{}.xml'.format(xsd_schema.split('/')[-1]),
            xsd_schema)

    # types/bodemclassificatie
    update_file(
        'types/bodemclassificatie/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'bodem:bodemclassificaties&maxFeatures=1&'
            'CQL_Filter=Bodemclassificatiefiche=%27' +
            build_dov_url('data/belgischebodemclassificatie/2018-000146%27')))

    update_file(
        'types/bodemclassificatie/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'bodem:bodemclassificaties&maxFeatures=1&'
            'CQL_Filter=Bodemclassificatiefiche=%27' +
            build_dov_url('data/belgischebodemclassificatie/2018-000146%27')),
        get_first_featuremember)

    update_file(
        'types/bodemclassificatie/fc_featurecatalogue.xml',
        build_dov_url('geonetwork/srv/dut/csw'
                      '?Service=CSW&Request=GetRecordById&Version=2.0.2'
                      '&outputSchema=http://www.isotc211.org/2005/gfc'
                      '&elementSetName=full&'
                      'id=9ace8d74-8daa-461f-86bb-273573bc7fa9'))

    update_file('types/bodemclassificatie/md_metadata.xml',
                build_dov_url(
                    'geonetwork/srv/dut/csw'
                    '?Service=CSW&Request=GetRecordById&Version=2.0.2'
                    '&outputSchema=http://www.isotc211.org/2005/gmd'
                    '&elementSetName=full&'
                    'id=7f668812-edc3-464b-870c-08e964f884b6'))

    update_file('types/bodemclassificatie/wfsdescribefeaturetype.xml',
                build_dov_url(
                    'geoserver/bodem/bodemclassificaties'
                    '/ows?service=wfs&version=1.1.0'
                    '&request=DescribeFeatureType'))

    for xsd_schema in Bodemclassificatie.get_xsd_schemas():
        update_file(
            'types/bodemclassificatie/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    # types/gw_vergunningen
    update_file(
        'types/grondwatervergunning/wfsgetfeature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'gw_vergunningen:alle_verg&maxFeatures=1&CQL_Filter'
            '=id=38598'))

    update_file(
        'types/grondwatervergunning/feature.xml',
        build_dov_url(
            'geoserver/ows?service=WFS'
            '&version=1.1.0&request=GetFeature&typeName='
            'gw_vergunningen:alle_verg&maxFeatures=1&CQL_Filter'
            '=id=38598'), get_first_featuremember)

    update_file(
        'types/grondwatervergunning/fc_featurecatalogue.xml',
        build_dov_url('geonetwork/srv/dut/csw'
                      '?Service=CSW&Request=GetRecordById&Version=2.0.2'
                      '&outputSchema=http://www.isotc211.org/2005/gfc'
                      '&elementSetName=full&id=5e605651-'
                      '6b2e-406f-863f-d2eda4d3e534'))

    update_file('types/grondwatervergunning/md_metadata.xml',
                build_dov_url(
                    'geonetwork/srv/dut/csw'
                    '?Service=CSW&Request=GetRecordById&Version=2.0.2'
                    '&outputSchema=http://www.isotc211.org/2005/gmd'
                    '&elementSetName=full&id=1b8d9dce-40b2-450c-81ed-'
                    'decb453b3143'))

    update_file('types/grondwatervergunning/wfsdescribefeaturetype.xml',
                build_dov_url(
                    'geoserver/gw_vergunningen/alle_verg'
                    '/ows?service=wfs&version=1.1.0'
                    '&request=DescribeFeatureType'))

    for xsd_schema in GrondwaterVergunning.get_xsd_schemas():
        update_file(
            'types/grondwatervergunning/xsd_%s.xml' %
            xsd_schema.split('/')[-1], xsd_schema)

    for r in pool.join():
        sys.stdout.write(r.get_result())



# -*- coding: utf-8 -*-
"""Module containing extra query classes to build attribute search queries."""

from owslib.fes import OgcExpression, Or, PropertyIsEqualTo


class PropertyInList(OgcExpression):
    """Filter expression to test whether a given property has one of the
    values from a list.

    Internally translates to an Or combination of PropertyIsEqualTo
    expressions:

    PropertyInList('methode', ['spade', 'spoelboring']) is equivalent to

    Or([PropertyIsEqualTo('methode', 'spade'), PropertyIsEqualTo('methode',
    'spoelboring')])

    """

    def __init__(self, propertyname, lst):
        """Initialisation.

        Parameters
        ----------
        propertyname : str
            Name of the attribute to query.
        lst : list of str
            List of literals to match against (exact matches).

        Raises
        ------
        ValueError
            If the given list does not contain at least a single item.

        """
        super(PropertyInList, self).__init__()

        if not isinstance(lst, list) and not isinstance(lst, set):
            raise ValueError('list should be of type "list" or "set"')

        if len(set(lst)) < 1:
            raise ValueError('list should contain at least a single item')
        elif len(set(lst)) == 1:
            self.query = PropertyIsEqualTo(propertyname, set(lst).pop())
        else:
            self.query = Or(
                [PropertyIsEqualTo(propertyname, i) for i in sorted(set(lst))])

    def toXML(self):
        """Return the XML representation of the PropertyInList query.

        Returns
        -------
        xml : etree.ElementTree
            XML representation of the PropertyInList

        """
        return self.query.toXML()


class Join(PropertyInList):
    """Filter expression to join different searches together.

    Internally translates to a PropertyInList:

    Join(df, 'pkey_boring') is equivalent to

    PropertyInList('pkey_boring', list(df['pkey_boring')) which is
    equivalent to

    Or([PropertyIsEqualTo('pkey_boring', x), PropertyIsEqualTo(
    'pkey_boring', y), ...]) for every x, y, in df['pkey_boring']

    """

    def __init__(self, dataframe, on, using=None):
        """Initialisation.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            Dataframe to use a basis for joining.
        on : str
            Name of the column in the queried datatype to join on.
        using : str, optional
            Name of the column in the dataframe to use for joining. By
            default, the same column name as in `on` is assumed.

        Raises
        ------
        ValueError
            If the `using` column is not present in the dataframe.

            If `using` is None and the `on` column is not present in the
            dataframe.

            If the dataframe does not contain at least a single non-null value
            in the `using` column.

        """
        if using is None:
            using = on

        if using not in list(dataframe):
            raise ValueError(
                "column '{}' should be present in the dataframe.".format(
                    using))

        value_list = list(dataframe[using].dropna().unique())

        if len(set(value_list)) < 1:
            raise ValueError("dataframe should contain at least a single "
                             "value in column '{}'.".format(using))

        super(Join, self).__init__(on, value_list)



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV groundwater samples."""
from ..types.grondwatermonster import GrondwaterMonster
from .abstract import AbstractSearch


class GrondwaterMonsterSearch(AbstractSearch):
    """Search class to retrieve information about groundwater samples
    (GrondwaterMonster).
    """

    def __init__(self, objecttype=GrondwaterMonster):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GrondwaterFilter type.
            Optional: defaults to the GrondwaterFilter type containing the
            fields described in the documentation.

        """
        super(GrondwaterMonsterSearch,
              self).__init__('gw_meetnetten:grondwatermonsters', objecttype)



import datetime
import gzip
import os
import sys
import tempfile
import time
from importlib import reload
from subprocess import Popen

import numpy as np
import pytest
from owslib.fes import PropertyIsEqualTo

import pydov
from pydov.search.boring import BoringSearch
from pydov.search.grondwaterfilter import GrondwaterFilterSearch
from pydov.util.caching import GzipTextFileCache
from pydov.util.dovutil import build_dov_url
from pydov.util.errors import XmlFetchWarning, XmlStaleWarning, XsdFetchWarning
from pydov.util.hooks import Hooks
from tests.abstract import ServiceCheck
from tests.test_util_hooks import HookCounter


@pytest.fixture(scope="module", autouse=True)
def dov_proxy_no_xdov():
    """Fixture to start the DOV proxy and set PYDOV_BASE_URL to route
    traffic through it.

    The DOV proxy behaves as the XDOV server would be unavailable.

    """
    process = Popen([sys.executable,
                     os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                  'stub', 'dov_proxy.py'),
                     '--dov-base-url', build_dov_url('/'),
                     '--no-xdov'])
    time.sleep(2)

    orig_base_url = os.environ.get('PYDOV_BASE_URL', None)
    os.environ['PYDOV_BASE_URL'] = 'http://localhost:1337/'

    yield

    if orig_base_url is not None:
        os.environ['PYDOV_BASE_URL'] = orig_base_url
    else:
        del(os.environ['PYDOV_BASE_URL'])

    process.terminate()
    process.communicate()


@pytest.fixture(scope="module", autouse=True)
def reload_modules(dov_proxy_no_xdov):
    """Reload the boring and grondwaterfilter modules after setting
    PYDOV_BASE_URL.

    These need to be reloaded because they use the PYDOV_BASE_URL at import
    time to set the location of XSD schemas.

    Parameters
    ----------
    dov_proxy_no_xdov : pytest.fixture
        Fixture starting the DOV proxy and setting PYDOV_BASE_URL accordingly.
    """
    reload(pydov.types.boring)
    reload(pydov.types.grondwaterfilter)

    yield

    reload(pydov.types.boring)
    reload(pydov.types.grondwaterfilter)


@pytest.fixture(scope="function", autouse=True)
def reset_cache(dov_proxy_no_xdov):
    """Reset the cache to a temporary folder to remove influence from other
    tests.

    The cache needs to be reset after setting the PYDOV_BASE_URL variable
    because at initialisation this URL is used to construct a regex for
    determining the datatype of an XML request.

    Parameters
    ----------
    dov_proxy_no_xdov : pytest.fixture
        Fixture starting the DOV proxy and setting PYDOV_BASE_URL accordingly.
    """
    gziptext_cache = GzipTextFileCache(
        cachedir=os.path.join(tempfile.gettempdir(), 'pydov_tests_error'),
        max_age=datetime.timedelta(seconds=0.1))
    gziptext_cache.remove()

    orig_cache = pydov.cache
    pydov.cache = gziptext_cache

    yield

    gziptext_cache.remove()
    pydov.cache = orig_cache


@pytest.fixture
def test_hook_count():
    """PyTest fixture temporarily disabling default hooks and installing
    HookCounter."""
    orig_hooks = pydov.hooks

    pydov.hooks = Hooks(
        (HookCounter(),)
    )
    yield

    pydov.hooks = orig_hooks


class TestNoXDOV(object):
    """Class grouping tests related failing DOV services."""

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_do_not_cache_error(self):
        """Test whether the 404 error page does not end up being cached."""
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        bs.search(query=PropertyIsEqualTo(
            'pkey_boring', build_dov_url('data/boring/2004-103984')))

        assert not os.path.exists(os.path.join(
            pydov.cache.cachedir, 'boring', '2004-103984.xml.gz'
        ))

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_do_not_overwrite_stale_cache(self):
        """Test whether a stale copy of the data which exists in the cache is
        not overwritten by the 404 error page."""
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        testdata_path = os.path.join(
            'tests', 'data', 'types', 'boring', 'boring.xml')

        cache_path = os.path.join(
            pydov.cache.cachedir, 'boring', '2004-103984.xml.gz'
        )
        os.makedirs(os.path.dirname(cache_path))

        with open(testdata_path, 'r') as testdata:
            with gzip.open(cache_path, 'wb') as cached_data:
                cached_data.write(testdata.read().encode('utf8'))
        time.sleep(0.5)

        bs.search(query=PropertyIsEqualTo(
            'pkey_boring', build_dov_url('data/boring/2004-103984')))

        with gzip.open(cache_path, 'rb') as cached_data:
            assert 'GEO-04/169-BNo-B1' in cached_data.read().decode('utf8')

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_stale_warning(self):
        """Test whether a stale version of the data from the cache is used in
        case of a service error, and if a warning is issued to the user."""
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        testdata_path = os.path.join(
            'tests', 'data', 'types', 'boring', 'boring.xml')

        cache_path = os.path.join(
            pydov.cache.cachedir, 'boring', '2004-103984.xml.gz'
        )
        os.makedirs(os.path.dirname(cache_path))

        with open(testdata_path, 'r') as testdata:
            with gzip.open(cache_path, 'wb') as cached_data:
                cached_data.write(testdata.read().encode('utf8'))
        time.sleep(0.5)

        with pytest.warns(XmlStaleWarning):
            df = bs.search(query=PropertyIsEqualTo(
                'pkey_boring', build_dov_url('data/boring/2004-103984')))

        assert not df.iloc[0].boorgatmeting
        assert df.iloc[0].boormethode == 'spade'

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_stale_disabled(self):
        """Test whether no stale version of the data from the cache is used
        when disabled, and if a warning is issued to the user."""
        pydov.cache.stale_on_error = False

        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        testdata_path = os.path.join(
            'tests', 'data', 'types', 'boring', 'boring.xml')

        cache_path = os.path.join(
            pydov.cache.cachedir, 'boring', '2004-103984.xml.gz'
        )
        os.makedirs(os.path.dirname(cache_path))

        with open(testdata_path, 'r') as testdata:
            with gzip.open(cache_path, 'wb') as cached_data:
                cached_data.write(testdata.read().encode('utf8'))
        time.sleep(0.5)

        with pytest.warns(XmlFetchWarning):
            df = bs.search(query=PropertyIsEqualTo(
                'pkey_boring', build_dov_url('data/boring/2004-103984')))

        assert np.isnan(df.iloc[0].boorgatmeting)
        assert np.isnan(df.iloc[0].boormethode)

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_wfs_data_present(self):
        """Test whether data available in the WFS is present in the dataframe
        in case of a service error in XDOV."""
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        df = bs.search(query=PropertyIsEqualTo(
            'pkey_boring', build_dov_url('data/boring/2016-122561')))

        assert df.iloc[0].gemeente == 'Wortegem-Petegem'

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_nan_and_fetch_warning(self):
        """Test whether the XML data is set tot NaN in case of an error and
        no stale cache is available. Also test if a warning is given to the
        user."""
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        with pytest.warns(XmlFetchWarning):
            df = bs.search(query=PropertyIsEqualTo(
                'pkey_boring', build_dov_url('data/boring/2016-122561')))

        assert np.isnan(df.iloc[0].boorgatmeting)

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_no_xsd_warning(self):
        """Test whether the metadata can still be retrieved, and that the
        XSD values are unavailable. Also test if a warning is given to the
        user."""
        with pytest.warns(XsdFetchWarning):
            gwf = GrondwaterFilterSearch(
                objecttype=pydov.types.grondwaterfilter.GrondwaterFilter)
            fields = gwf.get_fields()

        assert 'values' not in fields['aquifer_code']

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_no_xsd_wfs_only(self):
        """Test whether the WFS data is available, even if XSD schemas cannot
        be resolved."""
        gwf = GrondwaterFilterSearch(
            objecttype=pydov.types.grondwaterfilter.GrondwaterFilter)

        df = gwf.search(max_features=1)
        assert df.iloc[0].pkey_filter is not None

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_hooks_fetch_error(self, test_hook_count):
        """Test if the correct hooks are fired when the XML fails to be
        fetched from DOV.

        Parameters
        ----------
        test_hook_count : pytest.fixture
            Fixture removing default hooks and installing HookCounter.
        """
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        bs.search(query=PropertyIsEqualTo(
            'pkey_boring', build_dov_url('data/boring/2004-103984')))

        assert pydov.hooks[0].count_wfs_search_init == 1
        assert pydov.hooks[0].count_wfs_search_result == 1
        assert pydov.hooks[0].count_wfs_search_result_received == 1

        assert pydov.hooks[0].count_xml_received == 0
        assert pydov.hooks[0].count_xml_cache_hit == 0
        assert pydov.hooks[0].count_xml_downloaded == 0
        assert pydov.hooks[0].count_xml_stale_hit == 0
        assert pydov.hooks[0].count_xml_fetch_error == 1

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_hooks_stale(self, test_hook_count):
        """Test if the correct hooks are fired when a stale XML document is
        returned from the cache.

        Parameters
        ----------
        test_hook_count : pytest.fixture
            Fixture removing default hooks and installing HookCounter.
        """
        bs = BoringSearch(objecttype=pydov.types.boring.Boring)

        testdata_path = os.path.join(
            'tests', 'data', 'types', 'boring', 'boring.xml')

        cache_path = os.path.join(
            pydov.cache.cachedir, 'boring', '2004-103984.xml.gz'
        )
        os.makedirs(os.path.dirname(cache_path))

        with open(testdata_path, 'r') as testdata:
            with gzip.open(cache_path, 'wb') as cached_data:
                cached_data.write(testdata.read().encode('utf8'))
        time.sleep(0.5)

        bs.search(query=PropertyIsEqualTo(
            'pkey_boring', build_dov_url('data/boring/2004-103984')))

        assert pydov.hooks[0].count_wfs_search_init == 1
        assert pydov.hooks[0].count_wfs_search_result == 1
        assert pydov.hooks[0].count_wfs_search_result_received == 1

        assert pydov.hooks[0].count_xml_received == 0
        assert pydov.hooks[0].count_xml_cache_hit == 0
        assert pydov.hooks[0].count_xml_downloaded == 0
        assert pydov.hooks[0].count_xml_stale_hit == 1
        assert pydov.hooks[0].count_xml_fetch_error == 0

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0



"""Module grouping tests for the boring search module."""

import pytest

from pydov.types.bodemsite import Bodemsite
from pydov.types.bodemlocatie import Bodemlocatie
from pydov.types.bodemdiepteinterval import Bodemdiepteinterval
from pydov.types.bodemobservatie import Bodemobservatie
from pydov.types.bodemmonster import Bodemmonster
from pydov.types.bodemclassificatie import Bodemclassificatie
from pydov.types.boring import Boring
from pydov.types.fields import XmlField
from pydov.types.grondwaterfilter import GrondwaterFilter
from pydov.types.grondwatermonster import GrondwaterMonster
from pydov.types.grondmonster import Grondmonster
from pydov.types.interpretaties import (
    GecodeerdeLithologie,
    HydrogeologischeStratigrafie,
    InformeleStratigrafie,
    LithologischeBeschrijvingen,
    FormeleStratigrafie,
    GeotechnischeCodering,
    QuartairStratigrafie,
    InformeleHydrogeologischeStratigrafie,
)
from pydov.types.sondering import Sondering


type_objects = [Bodemsite,
                Bodemlocatie,
                Bodemdiepteinterval,
                Bodemobservatie,
                Bodemmonster,
                Bodemclassificatie,
                Boring,
                Sondering,
                GrondwaterFilter,
                GrondwaterMonster,
                InformeleStratigrafie,
                FormeleStratigrafie,
                HydrogeologischeStratigrafie,
                InformeleHydrogeologischeStratigrafie,
                GecodeerdeLithologie,
                LithologischeBeschrijvingen,
                GeotechnischeCodering,
                QuartairStratigrafie,
                Grondmonster]


@pytest.mark.parametrize("objecttype", type_objects)
def test_get_fields_sourcewfs(objecttype):
    """Test the get_fields method for fields of the WFS source.

    Test whether all returned fields have 'wfs' as their 'source'.

    """
    fields = objecttype.get_fields(source=('wfs',))
    for field in fields.values():
        assert field['source'] == 'wfs'


@pytest.mark.parametrize("objecttype", type_objects)
def test_get_fields_sourcexml(objecttype):
    """Test the Boring.get_fields method for fields of the XML source.

    Test whether all returned fields have 'xml' as their 'source'.

    """
    fields = objecttype.get_fields(source=('xml',))
    for field in fields.values():
        assert field['source'] == 'xml'


@pytest.mark.parametrize("objecttype", type_objects)
def test_get_fields_subtypes_notnull(objecttype):
    """Test the get_fields method for fields of the subtype.

    Test whether all returned fields have 'notnull' = False.

    Fields of a subtype cannot be mandatory since the subtype itself is not
    compulsory.

    """
    allfields = objecttype.get_field_names()
    ownfields = objecttype.get_field_names(include_subtypes=False)
    subfields = [f for f in allfields if f not in ownfields]

    fields = objecttype.get_fields(source=('xml',))
    for field in fields.values():
        if field['name'] in subfields:
            assert field['notnull'] is False


@pytest.mark.parametrize("objecttype", type_objects)
def test_extend_fields_no_extra(objecttype):
    """Test the extend_fields method for empty extra_fields.

    Test whether the returned fields match the existing fields.
    Test whether the returned fields are not the same fields as the original
    fields.

    """
    fields = objecttype.extend_fields([])
    assert fields == objecttype.fields
    assert fields is not objecttype.fields


@pytest.mark.parametrize("objecttype", type_objects)
def test_extend_fields_with_extra(objecttype):
    """Test the extend_fields method with extra_fields.

    Test whether the extra field is included.

    """
    extra_fields = [
        XmlField(name='grondwatersysteem',
                 source_xpath='/filter/ligging/grondwatersysteem',
                 definition='Grondwatersysteem waarin de filter hangt.',
                 datatype='string',
                 notnull=False)
    ]

    fields = objecttype.extend_fields(extra_fields)

    assert len(fields) == len(objecttype.fields) + len(extra_fields)

    assert fields[-1] == extra_fields[-1]



"""Module grouping tests for the pydov.util.caching module."""

import gzip
import os
import time

import pytest

from pydov.util.dovutil import build_dov_url


class TestPlainTextFileCacheCache(object):
    """Class grouping tests for the pydov.util.caching.PlainTextFileCache
    class."""

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_clean(self, plaintext_cache, mp_remote_xml):
        """Test the clean method.

        Test whether the cached file and the cache directory are nonexistent
        after the clean method has been called.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        plaintext_cache.clean()
        assert os.path.exists(cached_file)
        assert os.path.exists(plaintext_cache.cachedir)

        time.sleep(1.5)
        plaintext_cache.clean()
        assert not os.path.exists(cached_file)
        assert os.path.exists(plaintext_cache.cachedir)

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_remove(self, plaintext_cache, mp_remote_xml):
        """Test the remove method.

        Test whether the cache directory is nonexistent after the remove
        method has been called.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        plaintext_cache.remove()
        assert not os.path.exists(cached_file)
        assert not os.path.exists(plaintext_cache.cachedir)

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_get_save(self, plaintext_cache, mp_remote_xml):
        """Test the get method.

        Test whether the document is saved in the cache.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_get_reuse(self, plaintext_cache, mp_remote_xml):
        """Test the get method.

        Test whether the document is saved in the cache and reused in a
        second function call.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        first_download_time = os.path.getmtime(cached_file)

        time.sleep(0.5)
        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        # assure we didn't redownload the file:
        assert os.path.getmtime(cached_file) == first_download_time

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_get_invalid(self, plaintext_cache, mp_remote_xml):
        """Test the get method.

        Test whether the document is saved in the cache not reused if the
        second function call is after the maximum age of the cached file.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        first_download_time = os.path.getmtime(cached_file)

        time.sleep(1.5)
        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        # assure we did redownload the file, since original is invalid now:
        assert os.path.getmtime(cached_file) > first_download_time

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_save_content(self, plaintext_cache, mp_remote_xml):
        """Test whether the data is saved in the cache.

        Test if the contents of the saved document are the same as the
        original data.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        with open('tests/data/types/boring/boring.xml', 'r',
                  encoding='utf-8') as ref:
            ref_data = ref.read()

        with open(cached_file, 'r', encoding='utf-8') as cached:
            cached_data = cached.read()

        assert cached_data == ref_data

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_reuse_content(self, plaintext_cache, mp_remote_xml):
        """Test whether the saved data is reused.

        Test if the contents returned by the cache are the same as the
        original data.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        with open('tests/data/types/boring/boring.xml', 'r') as ref:
            ref_data = ref.read().encode('utf-8')

        cached_data = plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))

        assert cached_data == ref_data

    @pytest.mark.parametrize('plaintext_cache', [[]],
                             indirect=['plaintext_cache'])
    def test_return_type(self, plaintext_cache, mp_remote_xml):
        """The the return type of the get method.

        Test whether the get method returns the data in the same datatype (
        i.e. bytes) regardless of the data was cached or not.

        Parameters
        ----------
        plaintext_cache : pytest.fixture providing
                pydov.util.caching.PlainTextFileCache
            PlainTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            plaintext_cache.cachedir, 'boring', '2004-103984.xml')

        plaintext_cache.clean()
        assert not os.path.exists(cached_file)

        ref_data = plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert isinstance(ref_data, bytes)

        assert os.path.exists(cached_file)

        cached_data = plaintext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert isinstance(cached_data, bytes)


class TestGzipTextFileCacheCache(object):
    """Class grouping tests for the pydov.util.caching.PlainTextFileCache
    class."""

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_clean(self, gziptext_cache, mp_remote_xml):
        """Test the clean method.

        Test whether the cached file and the cache directory are nonexistent
        after the clean method has been called.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        gziptext_cache.clean()
        assert os.path.exists(cached_file)
        assert os.path.exists(gziptext_cache.cachedir)

        time.sleep(1.5)
        gziptext_cache.clean()
        assert not os.path.exists(cached_file)
        assert os.path.exists(gziptext_cache.cachedir)

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_remove(self, gziptext_cache, mp_remote_xml):
        """Test the remove method.

        Test whether the cache directory is nonexistent after the remove
        method has been called.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        gziptext_cache.remove()
        assert not os.path.exists(cached_file)
        assert not os.path.exists(gziptext_cache.cachedir)

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_get_save(self, gziptext_cache, mp_remote_xml):
        """Test the get method.

        Test whether the document is saved in the cache.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_get_reuse(self, gziptext_cache, mp_remote_xml):
        """Test the get method.

        Test whether the document is saved in the cache and reused in a
        second function call.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        first_download_time = os.path.getmtime(cached_file)

        time.sleep(0.5)
        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        # assure we didn't redownload the file:
        assert os.path.getmtime(cached_file) == first_download_time

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_get_invalid(self, gziptext_cache, mp_remote_xml):
        """Test the get method.

        Test whether the document is saved in the cache not reused if the
        second function call is after the maximum age of the cached file.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        first_download_time = os.path.getmtime(cached_file)

        time.sleep(1.5)
        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        # assure we did redownload the file, since original is invalid now:
        assert os.path.getmtime(cached_file) > first_download_time

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_save_content(self, gziptext_cache, mp_remote_xml):
        """Test whether the data is saved in the cache.

        Test if the contents of the saved document are the same as the
        original data.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        with open('tests/data/types/boring/boring.xml', 'r',
                  encoding='utf-8') as ref:
            ref_data = ref.read()

        with gzip.open(cached_file, 'rb') as cached:
            cached_data = cached.read().decode('utf-8')

        assert cached_data == ref_data

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_reuse_content(self, gziptext_cache, mp_remote_xml):
        """Test whether the saved data is reused.

        Test if the contents returned by the cache are the same as the
        original data.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert os.path.exists(cached_file)

        with open('tests/data/types/boring/boring.xml', 'r') as ref:
            ref_data = ref.read().encode('utf-8')

        cached_data = gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))

        assert cached_data == ref_data

    @pytest.mark.parametrize('gziptext_cache', [[]],
                             indirect=['gziptext_cache'])
    def test_return_type(self, gziptext_cache, mp_remote_xml):
        """The the return type of the get method.

        Test whether the get method returns the data in the same datatype (
        i.e. bytes) regardless of the data was cached or not.

        Parameters
        ----------
        gziptext_cache : pytest.fixture providing
                pydov.util.caching.GzipTextFileCache
            GzipTextFileCache using a temporary directory and a maximum age
            of 1 second.
        mp_remote_xml : pytest.fixture
            Monkeypatch the call to the remote DOV service returning an XML
            document.

        """
        cached_file = os.path.join(
            gziptext_cache.cachedir, 'boring', '2004-103984.xml.gz')

        gziptext_cache.clean()
        assert not os.path.exists(cached_file)

        ref_data = gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert isinstance(ref_data, bytes)

        assert os.path.exists(cached_file)

        cached_data = gziptext_cache.get(
            build_dov_url('data/boring/2004-103984.xml'))
        assert isinstance(cached_data, bytes)



"""Module grouping tests for the pydov.types.bodemclassificatie module."""

from pydov.types.bodemclassificatie import Bodemclassificatie
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/bodemclassificatie' \
    '/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemclassificatie/feature.xml'


class TestBodemclassificatie(AbstractTestTypes):
    """Class grouping tests for the
    pydov.types.bodemclassificatie.Bodemclassificatie class."""

    datatype_class = Bodemclassificatie
    namespace = 'https://www.dov.vlaanderen.be/bodem'
    pkey_base = build_dov_url('data/')

    field_names = [
        'pkey_bodemclassificatie', 'pkey_bodemlocatie', 'x', 'y', 'mv_mtaw',
        'classificatietype', 'bodemtype', 'auteurs'
    ]
    field_names_subtypes = []
    field_names_nosubtypes = [
        'pkey_bodemclassificatie', 'pkey_bodemlocatie', 'x', 'y', 'mv_mtaw',
        'classificatietype', 'bodemtype', 'auteurs'
    ]

    valid_returnfields = ('pkey_bodemclassificatie', 'bodemtype')
    valid_returnfields_subtype = ('pkey_bodemclassificatie', 'bodemtype')

    inexistent_field = 'onbestaand'

    def test_get_field_names_wrongreturnfields_nosubtypes(self):
        """Test the get_field_names method when disabling subtypes
        and including an otherwise valid return field.

        Test whether an InvalidFieldError is raised.

        Override because bodemclassificatie has no subtypes.
        """
        assert True

    def test_from_wfs_element(self, wfs_feature):
        """Test the from_wfs_element method.

        Test whether we can construct an instance from a WFS response element.

        Parameters
        ----------
        wfs_feature : pytest.fixture returning etree.Element
            Fixture providing an XML element representing a single record of
            the WFS layer.

        """
        feature = self.datatype_class.from_wfs_element(
            wfs_feature, self.namespace)

        assert isinstance(feature, self.datatype_class)

        if self.pkey_base is not None:
            assert feature.pkey.startswith(self.pkey_base)

        assert isinstance(feature.data, dict)
        assert isinstance(feature.subdata, dict)



# -*- coding: utf-8 -*-
"""Module grouping tests for the bodemmonster search module."""
import datetime

from owslib.fes import PropertyIsEqualTo

from pydov.search.bodemmonster import BodemmonsterSearch
from pydov.types.bodemmonster import Bodemmonster
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemmonster/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemmonster/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemmonster/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemmonster/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemmonster/feature.xml'
location_dov_xml = 'tests/data/types/bodemmonster/bodemmonster.xml'
location_xsd_base = 'tests/data/types/bodemmonster/xsd_*.xml'


class TestBodemmonsterSearch(AbstractTestSearch):

    search_instance = BodemmonsterSearch()
    datatype_class = Bodemmonster

    valid_query_single = PropertyIsEqualTo(propertyname='identificatie',
                                           literal='A0055738')

    inexistent_field = 'onbestaand'
    wfs_field = 'identificatie'
    xml_field = 'opmerking'

    valid_returnfields = ('pkey_bodemmonster', 'identificatie')
    valid_returnfields_subtype = (
        'pkey_bodemmonster',
        'identificatie',
        'tijdstip_monstername')
    valid_returnfields_extra = (
        'pkey_bodemmonster',
        'identificatie',
        'Opdrachten')

    df_default_columns = ['pkey_bodemmonster', 'pkey_bodemlocatie',
                          'pkey_parent',
                          'x', 'y', 'mv_mtaw', 'identificatie',
                          'datum_monstername', 'tijdstip_monstername',
                          'type', 'monstername_door', 'techniek',
                          'condities', 'diepte_van_cm', 'diepte_tot_cm',
                          'labo']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.datum_monstername.unique()[0] == datetime.date(2015, 6, 30)



# -*- coding: utf-8 -*-
"""Module grouping tests for the bodemobservatie search module."""
import datetime

from owslib.fes import PropertyIsEqualTo

from pydov.search.bodemobservatie import BodemobservatieSearch
from pydov.types.bodemobservatie import Bodemobservatie
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemobservatie/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemobservatie/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemobservatie/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemobservatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemobservatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemobservatie/bodemobservatie.xml'
location_xsd_base = 'tests/data/types/bodemobservatie/xsd_*.xml'


class TestBodemobservatieSearch(AbstractTestSearch):

    search_instance = BodemobservatieSearch()
    datatype_class = Bodemobservatie

    valid_query_single = PropertyIsEqualTo(propertyname='Bodemobservatie',
                                           literal='2019-1027345')

    inexistent_field = 'onbestaand'
    wfs_field = 'parameter'
    xml_field = 'observatiedatum'

    valid_returnfields = (
        'pkey_bodemobservatie',
        'parameter')
    valid_returnfields_subtype = (
        'pkey_bodemobservatie',
        'parameter',
        'observatiedatum')
    valid_returnfields_extra = (
        'pkey_bodemobservatie',
        'parameter',
        'Opmerkingen')

    df_default_columns = [
        'pkey_bodemobservatie', 'pkey_bodemlocatie',
        'pkey_parent', 'x', 'y', 'mv_mtaw', 'diepte_van_cm',
        'diepte_tot_cm', 'observatiedatum', 'invoerdatum',
        'parametergroep', 'parameter', 'detectie', 'waarde',
        'eenheid', 'veld_labo', 'methode', 'betrouwbaarheid',
        'fractiemeting_ondergrens', 'fractiemeting_bovengrens',
        'fractiemeting_waarde']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.observatiedatum.unique()[0] == datetime.date(2015, 12, 10)



"""Module grouping tests for the bodemclassificatie search module."""
from owslib.fes import PropertyIsEqualTo

from pydov.search.bodemclassificatie import BodemclassificatieSearch
from pydov.types.bodemclassificatie import Bodemclassificatie
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemclassificatie/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemclassificatie/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemclassificatie/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemclassificatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemclassificatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemclassificatie/bodemclassificatie.xml'
location_xsd_base = 'tests/data/types/bodemclassificatie/xsd_*.xml'


class TestBodemclassificatieSearch(AbstractTestSearch):

    search_instance = BodemclassificatieSearch()
    datatype_class = Bodemclassificatie

    valid_query_single = PropertyIsEqualTo(
        propertyname='pkey_bodemclassificatie',
        literal=build_dov_url('data/belgischebodemclassificatie/2018-000146'))

    inexistent_field = 'onbestaand'
    wfs_field = 'bodemtype'
    xml_field = None

    valid_returnfields = ('pkey_bodemclassificatie', 'bodemtype',
                          'classificatietype')
    valid_returnfields_subtype = None
    valid_returnfields_extra = None

    df_default_columns = [
        'pkey_bodemclassificatie', 'pkey_bodemlocatie', 'x', 'y', 'mv_mtaw',
        'classificatietype', 'bodemtype', 'auteurs'
    ]



import argparse

import requests
from flask import Flask, request

app = Flask(__name__)

dov_base_url = 'https://www.dov.vlaanderen.be/'
no_xdov = False

error_404 = """<!-- THEME DEBUG -->
<!-- THEME HOOK: 'html' -->
<!-- FILE NAME SUGGESTIONS:
* html--error--502.html.html.twig
* html--error.html.twig
x html.html.twig
-->
<!-- BEGIN OUTPUT from 'themes/dov_webuniversum/html.html.twig' -->
<html>
</html>
"""


def rewrite_content_reverse(content):
    """Rewrite content coming from DOV.

    Parameters
    ----------
    content : str or bytes
        The content to rewrite.

    Returns
    -------
    str or bytes
        The content where all URLs starting with DOV_BASE_URL are rewritten
        to localhost.
    """
    def rewrite(content):
        return content.replace(dov_base_url, 'http://localhost:1337/')

    if isinstance(content, bytes):
        return rewrite(content.decode('utf8')).encode('utf8')
    else:
        return rewrite(content)


def rewrite_content_forward(content):
    """Rewrite content going to DOV.

    Parameters
    ----------
    content : str or bytes
        The content to rewrite.

    Returns
    -------
    str or bytes
        The content where all URLs starting with DOV_BASE_URL are rewritten
        to localhost.
    """
    def rewrite(content):
        return content.replace('http://localhost:1337/', dov_base_url)

    if isinstance(content, bytes):
        return rewrite(content.decode('utf8')).encode('utf8')
    else:
        return rewrite(content)


@app.route('/', defaults={'path': ''}, methods=['HEAD'])
@app.route('/<path:path>', methods=['HEAD'])
def proxy_head(path):
    """Proxy HEAD requests.

    Parameters
    ----------
    path : str
        The path on the DOV server to proxy to.

    Returns
    -------
    flask.Response
        The response of the DOV service.
    """
    full_path = request.full_path
    r = requests.head(f'{dov_base_url.rstrip("/")}{full_path}')
    return r.content


@app.route('/', defaults={'path': ''}, methods=['GET'])
@app.route('/<path:path>', methods=['GET'])
def proxy_get(path):
    """Proxy GET requests.

    Parameters
    ----------
    path : str
        The path on the DOV server to proxy to.

    Returns
    -------
    flask.Response
        The response from the DOV service, content adjusted according to proxy
        rewrites.

        If --no-xdov was specified, returns 404 on all requests headed for
        XDOV.
    """
    full_path = request.full_path

    if no_xdov and (full_path.startswith('/data')
                    or full_path.startswith('/xdov')):
        return error_404, 404

    r = requests.get(f'{dov_base_url.rstrip("/")}{full_path}')
    return rewrite_content_reverse(r.content)


@app.route('/', defaults={'path': ''}, methods=['POST'])
@app.route('/<path:path>', methods=['POST'])
def proxy_post(path):
    """Proxy GET requests.

    Parameters
    ----------
    path : str
        The path on the DOV server to proxy to.

    Returns
    -------
    flask.Response
        The response from the DOV service, content adjusted according to proxy
        rewrites.
    """
    full_path = request.full_path
    data = rewrite_content_forward(request.data)
    r = requests.post(f'{dov_base_url.rstrip("/")}{full_path}', data)
    return rewrite_content_reverse(r.content)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='DOV proxy')

    parser.add_argument(
        '--dov-base-url', dest='dov_base_url',
        default='https://www.dov.vlaanderen.be/', action='store',
        help='DOV base URL to proxy to'
    )
    parser.add_argument(
        '--no-xdov', dest='no_xdov', action='store_true',
        help='XDOV requests return 404')

    args = parser.parse_args()

    no_xdov = args.no_xdov
    dov_base_url = args.dov_base_url

    app.run(host='127.0.0.1', port=1337)
